# vim:set et sts=4 sw=4:
#
# ibus - The Input Bus
#
# Copyright (c) 2009 Takao Fujiwara <takao.fujiwara1@gmail.com>
# Copyright (c) 2009 Red Hat, Inc.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330,
# Boston, MA  02111-1307  USA

import ibus
import gtk
import gtk.gdk as gdk
import gobject
import pango

try:
    from egg.trayicon import TrayIcon
except ImportError:
    TrayIcon = lambda a: None

LABEL_DESC_STR = "Sans 10"
LABEL_BG_COLOR = '#c9e1fc'
LABEL_FG_COLOR = '#707efb'

class StatusIcon(gobject.GObject):
    __gtype_name__ = "IBusStatusIcon"
    __gsignals__ = {
    "activate" :(
        gobject.SIGNAL_RUN_FIRST | gobject.SIGNAL_ACTION,
        gobject.TYPE_NONE,
       ()),
    "popup-menu" :(
        gobject.SIGNAL_RUN_FIRST | gobject.SIGNAL_ACTION,
        gobject.TYPE_NONE,
       (gobject.TYPE_UINT, gobject.TYPE_UINT)),
    }

    def __init__(self):
        self.__size = 24
        self.__tray_icon = None
        self.__label = None
        self.__current_im_name = None
        self.__pixbuf = None
        self.__image = None
        self.__image_data = None
        self.__storage_type = None
        self.__draw_label = False

        super(StatusIcon, self).__init__()
        self.__tray_icon = TrayIcon(self.__gtype_name__)
        self.__tray_icon.add_events(gtk.gdk.BUTTON_PRESS_MASK |
                                    gtk.gdk.BUTTON_RELEASE_MASK |
                                    gtk.gdk.SCROLL_MASK)
        self.__tray_icon.connect("button-press-event", self.__button_press_cb)
        self.__image = gtk.Image()
        self.__image.set_flags(gtk.CAN_FOCUS)
        self.__tray_icon.add(self.__image)
        self.__image.connect_after("expose-event", self.__expose_cb)
        self.__image.show()

    def __emit_activate_signal(self):
        self.emit("activate")

    def __emit_popup_menu_signal(self, button, activate_time):
        self.emit("popup-menu", button, activate_time)

    def __button_press_cb(self, widget, event):
        if event.button == 1 and event.type == gtk.gdk.BUTTON_PRESS:
            self.__emit_activate_signal()
            return True
        if event.button == 3 and event.type == gtk.gdk.BUTTON_PRESS:
            self.__emit_popup_menu_signal(event.button, event.time)
            return True
        return False

    def __expose_cb(self, widget, event):
        gc = widget.style.fg_gc[gtk.STATE_NORMAL]
        tray_icon_geometry = self.__tray_icon.window.get_geometry()

        if self.__pixbuf:
            width = self.__pixbuf.get_width()
            height = self.__pixbuf.get_height()
            x = (tray_icon_geometry[2] - width) / 2
            y = (tray_icon_geometry[3] - height) / 2
            widget.window.draw_pixbuf(gc, self.__pixbuf, 0, 0, x, y, -1, -1)
        if self.__label and self.__draw_label:
            label_layout = widget.create_pango_layout(self.__label.get_text())
            desc = pango.FontDescription(LABEL_DESC_STR)
            label_layout.set_font_description(desc)
            label_extents = label_layout.get_pixel_extents()
            x = 1
            y = max(tray_icon_geometry[3] - label_extents[0][1] \
                - label_extents[0][3], 0)
            pad = 3
            if y > pad:
                y -= pad
            gc = gtk.gdk.GC(widget.window)
            gc.set_rgb_fg_color(gtk.gdk.color_parse(LABEL_BG_COLOR))
            widget.window.draw_layout(gc, x , y, label_layout)
            widget.window.draw_layout(gc, x + 2, y + 2, label_layout)
            widget.window.draw_layout(gc, x, y + 2, label_layout)
            widget.window.draw_layout(gc, x + 2, y, label_layout)
            gc.set_rgb_fg_color(gtk.gdk.color_parse(LABEL_FG_COLOR))
            widget.window.draw_layout(gc, x + 1, y + 1, label_layout)

    def __update_image(self, change_image):
        if change_image and self.__storage_type == gtk.IMAGE_PIXBUF:
            pixbuf = self.__image_data;
            width = pixbuf.get_width()
            height = pixbuf.get_height()
            if width > self.__size:
                width = self.__size
            if height > self.__size:
                height = self.__size
            self.__pixbuf = pixbuf.scale_simple(width, height, gdk.INTERP_BILINEAR)
        if change_image and self.__storage_type == gtk.IMAGE_ICON_NAME:
            icon_name = self.__image_data
            theme = gtk.icon_theme_get_default()
            self.__pixbuf = theme.load_icon(icon_name, self.__size, 0)
        if self.__pixbuf:
            self.__image.set_from_pixbuf(self.__pixbuf)
            self.__image.queue_draw()

    def __set_image(self, storage_type, data, draw_label):
        self.__storage_type = storage_type
        self.__image_data = data
        self.__draw_label = draw_label 
        self.__update_image(not draw_label)

    def set_from_icon_name(self, icon_name, draw_label=True):
        if self.__tray_icon == None:
            super(StatusIcon, self).set_from_icon_name(icon_name)
            return
        if not icon_name:
            return
        self.__set_image(gtk.IMAGE_ICON_NAME, icon_name, draw_label)

    def set_from_pixbuf(self, pixbuf, draw_label=True):
        if self.__tray_icon == None:
            super(StatusIcon, self).set_from_pixbuf(pixbuf)
            return
        self.__set_image(gtk.IMAGE_PIXBUF, pixbuf, draw_label)

    def set_from_file(self, filename, draw_label=True):
        if self.__tray_icon == None:
            super(StatusIcon, self).set_from_file(filename)
            return
        if not filename:
            return
        pixbuf = gdk.pixbuf_new_from_file (filename)
        self.set_from_pixbuf(pixbuf, draw_label)

    def set_tooltip_text(self, text):
        self.__tray_icon.set_tooltip_text(text)

    def set_tooltip(self, text):
        self.set_tooltip_text(text)

    def set_visible(self, visible):
        if self.__tray_icon == None:
            super(StatusIcon, self).set_visible(visible)
            return
        if visible:
            self.__tray_icon.show()
        else:
            self.__tray_icon.hide()
            self.__tray_icon.unrealize()

    def set_label(self, label):
        self.__label = label
        self.__update_image(False)

    def set_im_name(self, name):
        if name and name != self.__current_im_name:
            self.__current_im_name = name
            label = ibus.Text(unicode(name.strip()[0:2].upper()))
            self.set_label(label)

    def position_menu(self, menu, data):
        status_icon = data
        tray_icon = status_icon.__tray_icon
        direction = tray_icon.get_direction()
        screen = tray_icon.get_screen()
        menu.set_screen(screen)
        monitor_num = screen.get_monitor_at_window(tray_icon.window)
        if monitor_num < 0:
            monitor_num = 0
        menu.set_monitor(monitor_num)
        monitor = screen.get_monitor_geometry(monitor_num)
        (x, y) = tray_icon.window.get_origin()
        menu_req = menu.size_request()
        orientation = gtk.ORIENTATION_HORIZONTAL
        if orientation == gtk.ORIENTATION_VERTICAL:
            width = 0
            height = tray_icon.allocation.height;
            xoffset = tray_icon.allocation.width;
            yoffset = 0
        else:
            width = tray_icon.allocation.width;
            height = 0
            xoffset = 0
            yoffset = tray_icon.allocation.height;
        if direction == gtk.TEXT_DIR_RTL:
            if x - (menu_req[0] - width) >= monitor.x:
                x -= menu_req[0] - width
            elif x + xoffset + menu_req[0] < monitor.x + monitor.width:
                x += xoffset
            elif monitor.x + monitor.width - (x + xoffset) < x:
                x -= menu_req[0] - width
            else:
                x += xoffset
        else:
            if x + xoffset + menu_req[0] < monitor.x + monitor.width:
                x += xoffset
            elif x - (menu_req[0] - width) >= monitor.x:
                x -= menu_req[0] - width
            elif monitor.x + monitor.width - (x + xoffset) > x:
                x += xoffset
            else:
                x -= menu_req[0] - width
        if y + yoffset + menu_req[1] < monitor.y + monitor.height:
            y += yoffset
        elif y - (menu_req[1] - height) >= monitor.y:
            y -= menu_req[1] - height
        elif monitor.y + monitor.height - (y + yoffset) > y:
            y += yoffset
        else:
            y -= menu_req[1] - height
        return (x, y, False)

